import logging
import struct
import threading
import typing

import gatt

from lego_wireless import signals
from lego_wireless.constants import CHARACTERISTIC_UUID
from lego_wireless.enums import HubAttachedIOEvent
from lego_wireless.enums import HubProperty
from lego_wireless.enums import HubPropertyOperation
from lego_wireless.enums import MessageType
from lego_wireless.hub_io import HubIO
from lego_wireless.hub_io import LEDLight
from lego_wireless.hub_io import RGBLight
from lego_wireless.hub_io import TrainMotor
from lego_wireless.messages import HubAttachedIO
from lego_wireless.messages import HubProperties
from lego_wireless.messages import message_classes

logger = logging.getLogger(__name__)

DEFAULT_NAME = "HUB NO.4"


class Hub(gatt.Device):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ports = {}
        self.hub_characteristic = None
        self._battery_level = None
        self._connected = False

    @property
    def battery_level(self):
        return self._battery_level

    @property
    def connected(self):
        return self._connected

    @property
    def train_motor(self):
        for hub_io in self.ports.values():
            if isinstance(hub_io, TrainMotor):
                return hub_io

    @property
    def led_light(self):
        for hub_io in self.ports.values():
            if isinstance(hub_io, LEDLight):
                return hub_io

    @property
    def rgb_light(self) -> typing.Optional[RGBLight]:
        for hub_io in self.ports.values():
            if isinstance(hub_io, RGBLight):
                return hub_io
        return None

    def connect_succeeded(self):
        super().connect_succeeded()
        self._connected = True
        self.manager.hubs.add(self)
        logger.info("[%s] Connected" % (self.mac_address))

    def connect_failed(self, error):
        super().connect_failed(error)
        logger.warning("[%s] Connection failed: %s" % (self.mac_address, str(error)))

    def disconnect_succeeded(self):
        super().disconnect_succeeded()
        self._connected = False
        signals.hub_disconnected.send(self)

    def characteristic_enable_notification_succeeded(self, *args, **kwargs):
        super().characteristic_enable_notification_succeeded(*args, **kwargs)
        logger.info(
            "[%s] Characteristic enable notification_succeeded: %s %s"
            % (self.mac_address, str(args), str(kwargs))
        )

    def characteristic_value_updated(self, characteristic, value):
        super().characteristic_value_updated(characteristic, value)
        logger.debug("[%s] Message received: %r", self.mac_address, value)
        message = self.parse_message(value)
        logger.debug("[%s] Parsed message received: %r", self.mac_address, message)
        print(message)
        if isinstance(message, HubAttachedIO):
            if message.event in (
                HubAttachedIOEvent.AttachedIO,
                HubAttachedIOEvent.AttachedVirtualIO,
            ):
                if message.io_type in HubIO.registry:
                    self.ports[message.port] = HubIO.registry[message.io_type](
                        self, message.port
                    )
                    logger.info("New IO on port %d: %s", message.port, message.io_type)
                    signals.hub_io_connected.send(self, hub_io=self.ports[message.port])
                    self.hub_io_connected(self.ports[message.port])
                else:
                    logger.warning(
                        "Unimplemented IOType on port %d: %s",
                        message.port,
                        message.io_type,
                    )
            else:
                if message.port in self.ports:
                    logger.debug(
                        "Removed IO on port %d: %s",
                        message.port,
                        self.ports[message.port].io_type,
                    )
                    signals.hub_io_disconnected.send(
                        self, hub_io=self.ports[message.port]
                    )
                    self.hub_io_disconnected(self.ports[message.port])
                    del self.ports[message.port]
        elif (
            isinstance(message, HubProperties)
            and message.property == HubProperty.BatteryVoltage
            and message.operation == HubPropertyOperation.Update
        ):
            self._battery_level = int(message.payload[0])
            logger.debug(
                "[%s] Hub battery level: %d%%", self.mac_address, self._battery_level
            )
            signals.hub_battery_level.send(self, battery_level=self._battery_level)
        else:
            logger.warning("Unexpected message: %s", message)

    def hub_io_connected(self, hub_io):
        pass

    def hub_io_disconnected(self, hub_io):
        pass

    def send_message(self, message):
        if hasattr(message, "to_bytes"):
            logger.info("Sending message: %r", message)
            message = message.to_bytes()

        assert isinstance(message, bytes)

        length = len(message) + 2
        message = bytes([length, 0x00]) + message
        logger.info("Sending serialized message: %r", message)
        self.hub_characteristic.write_value(message)

    def parse_message(self, message):
        length = message[0]
        if not len(message) == length:
            logger.warning("Unexpected message length: %r", message)
            return
        message_type = MessageType(message[2])
        try:
            message_cls = message_classes[message_type]
        except KeyError:
            logger.warning("Unexpected message type: %s %r", message_type, message)
        else:
            return message_cls.from_bytes(message[3:])

    @property
    def name(self) -> typing.Optional[str]:
        try:
            return self._name
        except AttributeError:
            name = self._properties.Get("org.bluez.Device1", "Name")
            self._name: typing.Optional[str] = name if name != DEFAULT_NAME else None
            return self._name

    @name.setter
    def name(self, value: typing.Optional[str]):
        if not value:
            self.send_message(
                HubProperties(
                    property=HubProperty.AdvertisingName,
                    operation=HubPropertyOperation.Reset,
                    payload=b"",
                )
            )
            self._name = None
        elif len(value.encode()) > 14:
            raise ValueError("name cannot be longer than 14 characters")
        else:
            self.send_message(
                HubProperties(
                    property=HubProperty.AdvertisingName,
                    operation=HubPropertyOperation.Set,
                    payload=value.encode(),
                )
            )
            self._name = value

    def services_resolved(self):
        logger.debug("Services resolved for %s", self.mac_address)
        super().services_resolved()

        for service in self.services:
            for characteristic in service.characteristics:
                if str(characteristic.uuid) == CHARACTERISTIC_UUID:

                    self.hub_characteristic = characteristic
                    self.hub_characteristic.enable_notifications()

                    self.send_message(
                        struct.pack(
                            "BBB",
                            MessageType.HubProperties,
                            HubProperty.BatteryVoltage,
                            HubPropertyOperation.EnableUpdates,
                        )
                    )
                    self.send_message(
                        struct.pack(
                            "BBB",
                            MessageType.HubProperties,
                            HubProperty.Button,
                            HubPropertyOperation.EnableUpdates,
                        )
                    )
                    signals.hub_connected.send(self)
                    return
        logger.debug("Device %s is not a LWP Hub", self.mac_address)

    def async_connect(self):
        threading.Thread(target=self.connect).run()

    def async_disconnect(self):
        threading.Thread(target=self.disconnect).run()



import atexit
import logging
import threading

from . import signals
from .manager import HubManager
from lego_wireless.hub_io import LEDLight
from lego_wireless.hub_io import TrainMotor

hubs_seen = set()


def hub_discovered(sender, hub):
    if hub not in hubs_seen:
        hubs_seen.add(hub)
        print("Connecting Hub IO, %s", hub.mac_address)
        signals.hub_io_connected.connect(hub_io_connected, sender=hub)
        hub.connect()


def hub_io_connected(sender, hub_io):
    print("Let's go!")
    if isinstance(hub_io, TrainMotor):
        hub_io.set_speed(100)
    if isinstance(hub_io, LEDLight):
        hub_io.set_brightness(100)


def main():
    hub_manager = HubManager("hci0")
    atexit.register(hub_manager.stop)

    signals.hub_discovered.connect(hub_discovered, sender=hub_manager)

    hub_manager_thread = threading.Thread(target=hub_manager.run)
    hub_manager_thread.start()
    hub_manager.start_discovery()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()



import gatt

from . import signals
from .constants import SERVICE_UUID
from .hub import Hub


class HubManager(gatt.DeviceManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hubs = set()

    def make_device(self, mac_address):
        return Hub(mac_address=mac_address, manager=self)

    def device_discovered(self, device: gatt.Device):
        super().device_discovered(device)
        if SERVICE_UUID in map(
            str, device._properties.Get("org.bluez.Device1", "UUIDs")
        ):
            signals.hub_discovered.send(self, hub=device)

    def start_discovery(self):
        return super().start_discovery([SERVICE_UUID])

    def stop(self):
        for hub in self.hubs:
            hub.disconnect()
        super().stop()



import enum
from gettext import gettext as _


class MessageType(enum.IntEnum):
    HubProperties = 0x01
    HubActions = 0x02
    HubAlerts = 0x03
    HubAttachedIO = 0x04
    ErrorMessage = 0x05
    PortOutput = 0x81


class OutputSubCommand(enum.IntEnum):
    StartPower = 0x01


class IOType(enum.IntEnum):
    Motor = 0x0001
    TrainMotor = 0x0002
    Button = 0x0005
    LEDLight = 0x0008
    Voltage = 0x0014
    Current = 0x0015
    PiezoTone = 0x0016
    RGBLight = 0x0017


class HubAttachedIOEvent(enum.IntEnum):
    DetachedIO = 0x00
    AttachedIO = 0x01
    AttachedVirtualIO = 0x02


class ColorNo(enum.IntEnum):
    Off = 0
    Magenta = 1
    Purple = 2
    Blue = 3
    Cyan = 4
    Turquoise = 5
    Green = 6
    Yellow = 7
    Orange = 8
    Red = 9
    White = 10


color_names = {
    ColorNo.Off: _("off"),
    ColorNo.Magenta: _("magenta"),
    ColorNo.Purple: _("purple"),
    ColorNo.Blue: _("blue"),
    ColorNo.Cyan: _("cyan"),
    ColorNo.Turquoise: _("turquoise"),
    ColorNo.Green: _("green"),
    ColorNo.Yellow: _("yellow"),
    ColorNo.Orange: _("orange"),
    ColorNo.Red: _("red"),
    ColorNo.White: _("white"),
}


class HubProperty(enum.IntEnum):
    AdvertisingName = 0x01
    Button = 0x02
    FWVersion = 0x03
    HWVersion = 0x04
    RSSI = 0x05
    BatteryVoltage = 0x06
    BatteryType = 0x07
    ManufacturerName = 0x08
    RadioFirmwareVersion = 0x09
    WirelessProtocolVersion = 0x0A
    SystemTypeID = 0x0B
    HWNetworkID = 0x0C
    PrimaryMACAddress = 0x0D
    SecondaryMACAddress = 0x0E
    HardwareNetworkFamily = 0x0F


class HubPropertyOperation(enum.IntEnum):
    Set = 0x01
    EnableUpdates = 0x02
    DisableUpdates = 0x03
    Reset = 0x04
    RequestUpdate = 0x05
    Update = 0x06


class Startup(enum.IntEnum):
    BufferIfNecessary = 0b0000
    ExecuteImmediately = 0b0001


class Completion(enum.IntEnum):
    NoAction = 0b0000
    CommandFeedback = 0b0001


class MotorControl(enum.IntEnum):
    Float = 0
    Brake = 127


class ErrorCode(enum.IntEnum):
    ACK = 0x01
    MACK = 0x02
    BufferOverflow = 0x03
    Timeout = 0x04
    CommandNotRecognized = 0x05
    InvalidUse = 0x06
    Overcurrent = 0x07
    InternalError = 0x08



from __future__ import annotations

import struct
import typing

from .enums import ColorNo
from .enums import IOType
from .enums import MessageType


class HubIOMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        if getattr(cls, "io_type", None):
            cls.registry[cls.io_type] = cls
        return cls


class HubIO(metaclass=HubIOMetaclass):
    io_type: IOType
    registry: typing.Dict[IOType, HubIO] = {}

    def __init__(self, hub, port):
        self.hub = hub
        self.port = port


class TrainMotor(HubIO):
    io_type = IOType.TrainMotor

    def set_speed(self, value):
        self.hub.send_message(
            struct.pack(
                "BBBBBBBB",
                MessageType.PortOutput,
                self.port,
                0x00,
                0x60,
                0x00,
                value,
                0x00,
                0x00,
            )
        )


class LEDLight(HubIO):
    io_type = IOType.LEDLight

    def set_brightness(self, value):
        self.hub.send_message(
            struct.pack(
                "BBBBBBBB",
                MessageType.PortOutput,
                self.port,
                0x00,
                0x60,
                0x00,
                value,
                0x00,
                0x00,
            )
        )


class Voltage(HubIO):
    io_type = IOType.Voltage


class RGBLight(HubIO):
    io_type = IOType.RGBLight

    def set_rgb_color_no(self, color_no: ColorNo):
        self.hub.send_message(
            struct.pack(
                "BBBBBB", MessageType.PortOutput, self.port, 0x01, 0x51, 0x00, color_no
            )
        )


class Current(HubIO):
    io_type = IOType.Current



import collections
import struct

from .enums import ErrorCode
from .enums import HubAttachedIOEvent
from .enums import HubProperty
from .enums import HubPropertyOperation
from .enums import IOType
from .enums import MessageType


class HubAttachedIO(
    collections.namedtuple("HubAttachedIO", ("port", "event", "io_type"))
):
    @classmethod
    def from_bytes(cls, value):
        port, event = struct.unpack("BB", value[:2])
        event = HubAttachedIOEvent(event)
        if event in (
            HubAttachedIOEvent.AttachedIO,
            HubAttachedIOEvent.AttachedVirtualIO,
        ):
            io_type = IOType(struct.unpack("<H", value[2:4])[0])
        else:
            io_type = None
        return cls(port=port, event=event, io_type=io_type)

    def __repr__(self):
        return f"{type(self).__name__}({self.port!r}, {self.event!r}, {self.io_type!r})"


class HubProperties(
    collections.namedtuple("HubProperties", ("property", "operation", "payload"))
):
    def to_bytes(self):
        return (
            struct.pack(
                "BBB", MessageType.HubProperties, self.property, self.operation,
            )
            + self.payload
        )

    @classmethod
    def from_bytes(cls, value):
        return cls(
            property=HubProperty(value[0]),
            operation=HubPropertyOperation(value[1]),
            payload=value[2:],
        )

    def __repr__(self):
        return f"{type(self).__name__}({self.property!r}, {self.operation!r}, {self.payload!r})"


class PortOutputMessage(
    collections.namedtuple(
        "PortOutputMessage", ("port", "startup", "completion", "sub_command")
    )
):
    def to_bytes(self):
        return (
            struct.pack(
                "BBB", MessageType.PortOutput, self.startup << 4 + self.completion,
            )
            + self.sub_command.to_bytes()
        )


class StartPowerSubCommand(
    collections.namedtuple("StartPowerSubCommand", ("start_power",))
):
    def to_bytes(self):
        return struct.pack("BB",)


class ErrorMessage(
    collections.namedtuple("ErrorMessage", ("command_type", "error_code"))
):
    @classmethod
    def from_bytes(cls, value):
        return cls(command_type=MessageType(value[0]), error_code=ErrorCode(value[1]),)

    def __repr__(self):
        return f"{type(self).__name__}({self.command_type!r}, {self.error_code!r})"


message_classes = {
    MessageType.HubProperties: HubProperties,  # 0x01
    # MessageType.HubActions: HubActions,  # 0x02
    # MessageType.HubAlerts: HubAlerts, # 0x03
    MessageType.HubAttachedIO: HubAttachedIO,  # 0x04
    MessageType.ErrorMessage: ErrorMessage,  # 0x05
}



from blinker import signal

hub_discovered = signal("hub-discovered")

hub_connected = signal("hub-connected")
hub_disconnected = signal("hub-disconnected")

hub_io_connected = signal("hub-io-connected")
hub_io_disconnected = signal("hub-io-disconnected")

hub_battery_level = signal("hub-battery-level")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lego_wireless']

package_data = \
{'': ['*']}

install_requires = \
['blinker', 'gatt']

setup_kwargs = {
    'name': 'lego-wireless',
    'version': '0.2.1',
    'description': 'Control Lego Powered Up devices over Bluetooth',
    'long_description': '# Lego Wireless Protocol for Python\n\nA spare-time implementation of the Lego Wireless Protocol for Python, to support a pet project.\n\nSee `lego_wireless/__main__.py` for a usage example.\n',
    'author': 'Alex Dutton',
    'author_email': 'python-lego-wireless-protocol@alexdutton.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alexsdutton/python-lego-wireless-protocol',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
